/*
 * database.cc - a C++ class for disk databases using gdbm
 *
 */

#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <iostream>
#include <iomanip>
#include <assert.h>

#include "database.h"

using namespace std;



/*
 * If we include a file defining DATABASE_DEBUG, debugging output is printed
 * whenever a database entry is read or written.
 *
 */

#include "config.h"



database::database (char *s, char *d, time_t stamp) {
	struct stat fs;

	if (d) {
		if (*d!='~') {
			assert (dir = new char[strlen(d)+1]);
			strcpy (dir, d);
		} else {
			// Replace ~ by home directory
			assert (name = getenv ("HOME"));
			assert (dir = new char[strlen(name)+strlen(d)+1]);
			strcpy (dir,name);
			dir[strlen(name)] = '/';
			strcpy (&dir[strlen(name)+1],&d[1]);
		}

		// Check the directory and create it if necessary
		if (stat (dir,&fs)) {
			assert (!mkdir (dir,0777));
			assert (!stat (dir,&fs));
		};
		assert (S_ISDIR (fs.st_mode));
	} else {
		dir = (char*)".";
	};

	// Set the name including the directory
	assert (s);
	assert (name = new char[strlen(dir)+strlen(s)+2]);
	strcpy (name,dir);
	name[strlen(dir)] = '/';
	strcpy (&name[strlen(dir)+1],s);

	// Delete the file if it is too old
	#ifdef CHECK_TIME
	if ((!stat (name,&fs)) && (stamp>fs.st_mtime)) rm ();
	#endif

	op = 0;
	key.dptr = key_space;
	value.dptr = value_space;
}



bool database::open ()
{
	if (op) return 1;
	dbm = gdbm_open (name, 512, GDBM_WRCREAT|GDBM_FAST,
		S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH, NULL);
	if (dbm) op = 1;
	return op;
}



bool database::close ()
{
	if (!op) return 1;
	gdbm_close (dbm);
	op = 0;
	return 1;
}



bool database::rm ()
{
	char *com;

	if ( ! close () ) return 0;

	assert (com = new char[strlen(name)+8]);
	strcpy (com,"rm -rf ");
	strcpy (&com[7],name);
	if ( ! system (com) ) {
		delete [] com;
		return 0;
	};
	delete [] com;
	return 1;
}



// Byte-wise access to key and value

void database::set_key_bytes (char *s, int n)
{
	char *p;
	int i;

	// Allocate more memory if necessary
	if (key.dsize+n>key_alloc) {
		key_alloc = (key.dsize+n+255) & -256;
		assert (p = new char[key_alloc]);
		for (i=0; i<key.dsize; i++) p[i] = key.dptr[i];
		if (key.dptr != key_space) delete [] key.dptr;
		key.dptr = p;
	}

	// Copy the string to the key
	for (i=0; i<n; i++) key.dptr[key.dsize++] = *(s++);
}



void database::set_value_bytes (char *s, int n)
{
	char *p;
	int i;

	// Allocate more memory if necessary
	if (value.dsize+n>value_alloc) {
		value_alloc = (value.dsize+n+255) & -256;
		assert (p = new char[value_alloc]);
		for (i=0; i<value.dsize; i++) p[i] = value.dptr[i];
		if (value.dptr != value_space) delete [] value.dptr;
		value.dptr = p;
	}

	// Copy the string to the value
	for (i=0; i<n; i++) value.dptr[value.dsize++] = *(s++);
}



bool database::get_value_bytes (char *& s, int n)
{
	if (value_read+n>value.dsize) return 0;

	s = &value.dptr[value_read];
	value_read += n;

	return 1;
}



// Bit-wise access to key and value

void database::set_key_bits (int n, int bits)
{
	int i;

	// Fill the last bits of the current byte
	if ( i = (bits<key_pos ? bits : key_pos) ) { // the number of bits
		key_pos-=i; bits-=i;
		key_cur |= ( ( (n>>bits) & ((1<<i)-1) ) << key_pos );
		if (!key_pos) {
			set_key_bytes (&key_cur,1);
			key_cur = 0;
			key_pos = 0;
		}
		if (!bits) return; // all bits written
	}

	// Full bytes in the middle
	while (bits>=8) {
		bits-=8;
		key_cur = (n>>bits)&255;
		set_key_bytes (&key_cur,1);
		key_cur = 0;
	}

	// The remaining bits
	if (bits) {
		key_pos = 8-bits;
		key_cur = ( (n & ((1<<bits)-1) ) << key_pos);
	}
}



void database::set_value_bits (int n, int bits)
{
	int i;

	// Fill the last bits of the current byte
	if ( i = (bits<value_pos ? bits : value_pos) ) { // the number of bits
		value_pos-=i; bits-=i;
		value_cur |= ( ( (n>>bits) & ((1<<i)-1) ) << value_pos );
		if (!value_pos) {
			set_value_bytes (&value_cur,1);
			value_cur = 0;
			value_pos = 0;
		}
		if (!bits) return; // all bits written
	}

	// Full bytes in the middle
	while (bits>=8) {
		bits-=8;
		value_cur = (n>>bits)&255;
		set_value_bytes (&value_cur,1);
		value_cur = 0;
	}

	// The remaining bits
	if (bits) {
		value_pos = 8-bits;
		value_cur = ( (n & ((1<<bits)-1) ) << value_pos);
	}
}



bool database::get_value_bits (int& n, int bits)
{
	n = 0;
	int i;
	char *s;

	// Read the remaining bits of the current byte
	if ( i = (bits<value_pos ? bits : value_pos) ) { // the number of bits
		value_pos-=i; bits-=i;
		n = ( (int(value_cur)>>value_pos) & ((1<<i)-1) ) << bits;
		if (!bits) return 1;
	}

	// Full bytes in the middle
	while (bits>=8) {
		bits-=8;
		if ( ! get_value_bytes (s,1) ) return 0;
		value_cur = *s;
		n |= (int(value_cur) & 255) << bits;
	}

	// The remaining bits
	if (bits) {
		if ( ! get_value_bytes (s,1) ) return 0;
		value_cur = *s;
		value_pos = 8-bits;
		n |= ((int(value_cur)>>value_pos) & ((1<<bits)-1));
	}

	return 1;
}



// Write and read

#ifdef DATABASE_DEBUG

void write_database_debug (datum k)
{
	for (int i = 0; i<k.dsize; i++) {
		cout << " ";
		cout << int((unsigned char)k.dptr[i]);
	}
	cout << endl;
}

#endif



bool database::write ()
{
	if (!op) if ( ! open () ) return 0;
	gdbm_store (dbm, key, value, GDBM_REPLACE);

	#ifdef DATABASE_DEBUG
	cout << "*database* writing entry" << endl;
	cout << "  -> key  "; write_database_debug (key);
	cout << "  -> value"; write_database_debug (value);
	#endif

	return 1;
}



bool database::read ()
{
	if (!op) if ( ! open () ) return 0;

	#ifdef DATABASE_DEBUG
	cout << "*database* reading entry" << endl;
	cout << "  -> key  "; write_database_debug (key);
	#endif

	if (value.dptr != value_space) delete [] value.dptr;
	value = gdbm_fetch (dbm, key);
	if ( !value.dptr ) {
		#ifdef DATABASE_DEBUG
		cout << "  <- value UNDEFINED" << endl;
		#endif
		return 0;
	}
	value_read = 0;
	value_pos = 0;
	#ifdef DATABASE_DEBUG
	cout << "  <- value"; write_database_debug (value);
	#endif
	return 1;
}



// set/get_key/value functions for more data types

// strings

void set_key (database& db, char *s)
{
	db.set_key_bytes (s,strlen(s)+1);
}



void set_value (database& db, char *s)
{
	db.set_value_bytes (s,strlen(s)+1);
}



bool get_value (database& db, char *&s)
{
	if ( db.get_value_bytes (s,0) ) return 0;
	db.value_read += strlen(s)+1;

	return 1;
}



// chars

void set_key (database& db, char c)
{
	db.set_key_bytes (&c,1);
}



void set_value (database& db, char c)
{
	db.set_value_bytes (&c,1);
}



bool get_value (database& db, char& c)
{
	char *s;
	if (db.get_value_bytes (s,1)) return 0;
	c = *s;

	return 1;
}



/*
 * ints of any bit length. These ints are stored as bit fields of the following
 * form, from left to right:
 *  a) 1 1 ... 1 1 0 (where we have n 1's, and n+1 is the number of bytes in
 *     the bit field)
 *  b) 1 bit sign
 *  c) remaining bits for the value, with the most significant bits first
 *     (machine-independent). The number of bits is determined by n from (a).
 * This way, integers from -63 to 63 need only one byte of storage. In general,
 * we need one byte of storage for every 7 bits in the integer. When storing
 * numbers, n is chosen automaticaly to be the smallest value so that the
 * number fits.
 *
 * sizeof(int) doesn't matter.
 *
 */

// How many dual digits does the integer n have?

static inline int numbits (int n)
{
	int bi = 1;
	int i;

	if (n<0) return sizeof(int)>>3;

	while (n & -256) {n>>=8; bi+=8;};
	if (n & -16) {n>>=4; bi+=4;};
	if (n & -4) {n>>=2; bi+=2;};
	if (n & -2) bi+=1;

	return bi;
}



void set_key (database& db, int n)
{
	int sign = 0;
	char c;

	if (n<0) {n = -n; sign = 1;};

	// A fast function for small numbers...
	if (n<64) {
		if (sign) c = n+64;
		else c = n;
		db.set_key_bytes (&c,1);
		return;
	}

	int bytes = numbits(n)/7+1;
	db.set_key_bits ((1<<bytes)-2,bytes);
	db.set_key_bits (sign,1);
	db.set_key_bits (n,7*bytes-1);
}



void set_value (database& db, int n)
{
	int sign = 0;
	char c;

	if (n<0) {n = -n; sign = 1;};

	// A fast function for small numbers...
	if (n<64) {
		if (sign) c = n+64;
		else c = n;
		db.set_value_bytes (&c,1);
		return;
	}

	int bytes = numbits(n)/7+1;
	db.set_value_bits ((1<<bytes)-2,bytes);
	db.set_value_bits (sign,1);
	db.set_value_bits (n,7*bytes-1);
}



bool get_value (database& db, int& n)
{
	int bytes = 0;
	int sign;
	int bt;

	do {
		if ( ! db.get_value_bits (bt,1) ) return 0;
		bytes++;
	} while (bt);

	if ( ! db.get_value_bits (sign,1) ) return 0;
	if ( ! db.get_value_bits (n,7*bytes-1) ) return 0;
	if (sign) n = -n;
	return 1;
}

