/*
 * compute.cc - the main computation function
 *
 */

#include <stdlib.h>
#include <iostream>
#include <assert.h>

#include "compute.h"

/*
 * If we include a file defining INV_DEBUG, every invariant passed to compute()
 * is checked if it has valid entries. Normally we don't have to do this, as
 * checking of the user input is done in GW, and the recursive algorithm should
 * only require valid invariants.
 *
 * If we include a file setting COMPUTE_DEBUG to a positive integer, debugging
 * output is printed for all computations of invariants up to this level. This
 * shows how the invariants are computed.
 * 
 */

#include "config.h"



// The database to store the invariants. It's opened automatically when
// used for the first time.

#ifdef CHECK_TIME
database GWdbm((char*)"GW.dbm", (char*)DB_DIR, TIMESTAMP);
#else
database GWdbm((char*)"GW.dbm", (char*)DB_DIR);
#endif



// Compute an invariant

rational compute (const inv& x, int lvl)
{
	rational result;

	// Make a copy of the invariant
	inv iv(x);

	#ifdef COMPUTE_DEBUG
	if (lvl<=COMPUTE_DEBUG)
		cout << "*compute* entering level " << lvl
		     << " for " << iv << endl;
	#endif

	// Check whether the invariant is valid
	#ifdef INV_DEBUG
	iv.check ();
	#endif

	// Check if we can make some simplifications
	result = simplify (iv,lvl);
	if ( ! is_undefined (result) ) goto DONE;

	// Check if we can compute the invariant by splitting it into two
	// components
	result = split2 (iv,lvl);
	if ( ! is_undefined (result) ) goto DONE;

	// Check if we can compute the invariant using the topological
	// recursion relations
	result = trr_genusg (iv,lvl);
	if ( ! is_undefined (result) ) goto DONE;

	// Check if we can compute the invariant using the Virasoro conditions
	result = virasoro (iv,lvl);
	if ( ! is_undefined (result) ) goto DONE;

	// Check if we can compute the invariant by relative splitting
	result = relsplit (iv,lvl);
	if ( ! is_undefined (result) ) goto DONE;

	// We don't know how to compute the invariant - return UNDEFINED
	#ifdef COMPUTE_DEBUG
	if (lvl<=COMPUTE_DEBUG)
		cout << "*compute* leaving level " << lvl
		     << ", " << iv << " is unknown" << endl;
	#else
	cout << "*error* don't know how to compute " << iv << endl;
	#endif

	return rational ();

DONE:	
	#ifdef COMPUTE_DEBUG
	if (lvl<=COMPUTE_DEBUG)
		cout << "*compute* leaving level " << lvl
		     << ", " << iv << " is " << result << endl;
	#endif
	
	return result;
};

