/*
 * gwinv.cc - the GW function and the class used as argument for it
 *
 */

#include <stdlib.h>
#include <iostream>
#include <assert.h>

#include "gwinv.h"
#include "config.h"

// The fixed variables for degree, genus, dimension, ...

gwvar N('N',0); // dimension of the variety
gwvar E('E',1); // degree of the variety
gwvar G('G',0); // genus of the curves
gwvar D('D',0); // degree of the curves

// The cohomology classes

var H('H');
var psi('c');
var pc (int n) {return var ('P',n);};



gwlist operator+ (const gwlist& l1, const gwlist& l2)
{
	int i;

	gwlist result;
	result.alloc = l1.alloc + l2.alloc;
	if (!result.alloc) return result;

	assert (result.data = new gwentry[result.alloc]);
	for (i=0; i<l1.alloc; i++) result.data[i] = l1.data[i];
	for (i=0; i<l2.alloc; i++) result.data[i+l1.alloc] = l2.data[i];

	return result;
}



/*
 * Translate a monomial into a class at a point. We write the class into
 * the point iv(n) (space must have been allocated before).
 *
 * We return 0 if the class must give a zero contribution because
 * - it has a too big H power, or
 * - it contains a product of different primitive classes, or
 * - it contains a product of a primitive class with too many H's, or
 * - it contains a product of more than 2 primitive classes.
 * In this case, iv(n) is set to some undefined value. We return -1 if the
 * monomial contains invalid variables. If the class is OK and occurs with
 * a factor of 1/d, we return d (this happens as e.g. a product of two same
 * primitive classes is 1/d times H^top, where d is the degree of the variety).
 *
 * We have to pass a flag whether the invariant is relative. The entry mult
 * of iv(i) must have been set already. It will not be changed by this
 * function.
 *
 */

int set_point (inv& iv, int n, monomial m, int rel)
{
	int Hv = var_num ('H');
	int psiv = var_num ('c');
	int Pv = var_num ('P');
	int cof = 1; // the resulting coefficient
	int i;

	iv.pt[n].Hpow = 0;
	iv.pt[n].psipow = 0;
	for (i=0; i<m.vars(); i++) {
		if ( m.vr(i) == Hv ) {
			if ( iv.pt[n].Hpow<0 ) return 0; // primitive*H
			iv.pt[n].Hpow += m.ex(i);
			continue;
		}
		if ( m.vr(i) == psiv ) {
			iv.pt[n].psipow += m.ex(i);
			continue;
		}
		if ( (m.vr(i) & 63) == Pv && (m.vr(i)>>6) ) {
			if ( iv.xdeg==1 ) return -1; // primitive classes
				// occur for hypersurfaces of degree >1 only
			if (iv.pt[n].Hpow<0 || m.ex(i)>2) return 0;
			if ( m.ex(i)==1 ) {
				if (iv.pt[n].Hpow) return 0;
				iv.pt[n].Hpow = -(m.vr(i)>>6);
			} else {
				iv.pt[n].Hpow += iv.xdim;
				cof = iv.xdeg;
			}
			continue;
		}
		return -1; // invalid variable
	}

	// Check for too big H power
	if ( rel && !iv.pt[n].mult ) {
		if (iv.pt[n].Hpow > iv.xdim+1) return 0;
	} else {
		if (iv.pt[n].Hpow > iv.xdim) return 0;
	}

	return cof;
}



// The GW function.

rational GW (const gwlist& li)
{
	poly **py = NULL;

	int pts; // the number of marked points
	int vd = 0; // the virtual dimension
	int rel = 0; // do we have a relative invariant?
	int tm = 0; // the total relative multiplicity
	int cl = 0; // the number of classes with return 0 from set_point.
		    // As long as this number is positive, we don't get any
		    // contribution.
	int *primcof = NULL; // For every point, the coefficient coming from
			     // powers of two primitive classes
	int i,j;
	rational result = 0;
	rational coeff = 1;

	// Strip the irrelevant parts from the list, i.e. keep only the POINT
	// types. In theory, we could evaluate the VARSET entries too to
	// determine which special variables have been set explicitly in the
	// GW call. We don't do this here though. This could be used for later
	// extensions that would require gwentry possibilities like e.g. "D>=2"
	// instead of simple assignments "D=2".
	gwlist list;
	if (li.alloc) assert (list.data = new gwentry[li.alloc]);
	i = 0; pts = 0;
	for (i=0; i<li.alloc; i++) if (li.data[i].type==POINT)
		list.data[pts++] = li.data[i];
	list.alloc = pts;

	// Create a basic inv class, with only the classes missing.
	inv iv(pts);
	assert (primcof = new int[pts]);
	iv.points = pts;
	if ( (iv.xdim = N.value) < 0) {
		cout << "*error* dimension N=" << iv.xdim
		     << " invalid" << endl;
		result = rational ();
		goto DONE;
	}
	if ((iv.xdeg = E.value) < 1) {
		cout << "*error* degree E=" << iv.xdeg
		     << " invalid" << endl;
		result = rational ();
		goto DONE;
	}
	iv.cgenus = G.value;
	#ifndef NEGATIVE_GENUS
	if ( iv.cgenus < 0 ) {
		cout << "*error* genus G=" << iv.cgenus
		     << " invalid" << endl;
		result = rational ();
		goto DONE;
	}
	#endif
	if ((iv.cdeg=D.value)<0) {
		cout << "*error* degree D=" << iv.cdeg
		     << " invalid" << endl;
		result = rational ();
		goto DONE;
	}

	// Marked point initializations...
	// a) Define *py[i] to be list.data[i].cls for simplicity
	// b) Reset all polynomials
	// c) Compute the initial coefficient (and check that the poly is !=0)
	// d) Set the relative multiplicities
	if (pts) py = new poly *[pts];
	for (i=0; i<pts; i++) {
		py[i] = &(list.data[i].cls);
		py[i]->mon_first ();
		if ( ! py[i]->mon_next () ) goto DONE;
		coeff *= py[i]->mon_coeff ();
		tm += (iv.pt[i].mult = list.data[i].mult);
		if ( iv.pt[i].mult<0 ) {
			cout << "*error* negative multiplicity" << endl;
			result = rational ();
			goto DONE;
		}
		if ( iv.pt[i].mult ) rel = 1;
	};
	if ( (iv.cdeg) && (!iv.xdim) && (!rel) ) {
		cout << "*error* degree D=" << iv.cdeg
		     << " invalid" << endl;
		result = rational ();
		goto DONE;
	}

	// Set the initial H and psi powers, and check that there are no
	// invalid variables. Set the initial dimension of the invariant.
	for (i=0; i<pts; i++) {
		primcof[i] = j = set_point (iv,i,py[i]->mon_vars(),rel);
		if (j<0) {
			cout << "*error* polynomial " << *py[i]
			     << " contains invalid variables" << endl;
			result = rational ();
			goto DONE;
		}
		if (j>1) coeff /= j;
		if (j) vd -= iv.pt[i].vdim(iv.xdim);
		else cl++;
	}

	// The virtual dimension. Note that relative invariants are in fact
	// invariants of the ambient P^(n+1).
	if (rel) {
		if (iv.cgenus<0) { 
			cout << "*error* relative negative genus invariant"
			     << endl;
			result = rational ();
			goto DONE;
		}
		vd += 2*((2+iv.xdim)*iv.cdeg+(iv.xdim-2)*(1-iv.cgenus));
	} else {
		if (iv.cgenus<0) {
			vd += 2*((2+iv.xdim-iv.xdeg)*iv.cdeg +
					iv.xdim-3+iv.cgenus);
		} else {
			vd += 2*((2+iv.xdim-iv.xdeg)*iv.cdeg +
					(iv.xdim-3)*(1-iv.cgenus));
		}
	}

	// Check the total multiplicity
	if ( tm>iv.xdeg*iv.cdeg) {
		cout << "*error* invariant " << iv << endl
		     << "        has total multiplicity " << tm << "," << endl
		     << "        maximum is " << iv.xdeg*iv.cdeg << endl;
		result = rational ();
		goto DONE;
	}

	// A negative genus invariant must have at least 2 marked points
	#ifdef NEGATIVE_GENUS
	if ((iv.cgenus<0) && (pts<2)) {
		cout << "*error* invariant " << iv
		     << " must have at least 2 marked points" << endl;
		result = rational ();
		goto DONE;
	}
	#endif

	// Loop over all summands of all polynomials.
	do {
		if (!cl) { // otherwise one of the classes is trivially zero
			if (vd) {
				// wrong dimension
				cout << "*error* invariant " << iv
				     << " has (real) dimension " << vd << endl;
				result = rational ();
				goto DONE;
			}
			#ifdef NEGATIVE_GENUS
			if ( (iv.cgenus<0) && iv.pt[1].psipow ) {
				cout << "*error* invariant " << iv
				     << " has psi class on second point"
				     << endl;
				result = rational ();
				goto DONE;
			}
			#endif

			if ( is_undefined (result += coeff * compute (iv,1)) )
			goto DONE; // couldn't compute the invariant
		}

		// Go to the next combination of monomials.
		for (i=0; i<pts; i++) {
			// Reset old class i.
			coeff /= py[i]->mon_coeff ();
			if (primcof[i]>1) coeff *= primcof[i];
			if (primcof[i]) vd += iv.pt[i].vdim(iv.xdim);
			else cl--;

			j = 1;
			if ( ! py[i]->mon_next () ) {
				// If there are no more terms in polynomial i,
				// reset it, and keep in mind that we did so.
				py[i]->mon_first();
				py[i]->mon_next();
				j = 0;
			}

			// Initialize new class i.
			coeff *= py[i]->mon_coeff ();
			primcof[i] = set_point
				(iv,i,py[i]->mon_vars(),rel);
			if (primcof[i]>1) coeff /= primcof[i];
			if (primcof[i]) vd -= iv.pt[i].vdim(iv.xdim);
			else cl++;

			if (j) break;
		};
	} while (i<pts);

DONE:	delete [] py;
	delete [] primcof;
	return result;
}

