/*
 * split2.cc - functions for computing invariants by splitting them into
 *             two components
 *
 */

#include <stdlib.h>
#include <iostream>
#include <assert.h>

#include "split2.h"

/*
 * If we include a file setting SPLIT2_DEBUG to a positive integer, debugging
 * output is printed for all computations in this file of invariants up to
 * this level. This shows how the invariants are computed.
 * 
 */

#include "config.h"



#define NN iv.xdim
#define EE iv.xdeg
#define GG iv.cgenus
#define DD iv.cdeg



// Debugging macros

#define DEBUG_BEFORE(TEXT)						\
if (lvl<=SPLIT2_DEBUG)							\
	cout << "*split2* level " << lvl				\
	     << ": trying " TEXT << endl;
#define DEBUG_AFTER							\
if (lvl<=SPLIT2_DEBUG)							\
	cout << "*split2* level " << lvl				\
	     << ": result is " << result << endl;

// An integer division function that really rounds down. We assume that n
// is positive, but m can be negative.

int idiv (int m, int n) {
	return ( m>=0 ? m/n : -((n-m-1)/n));
}



/*
 * Compute all splittings of cohomology classes onto 2 components iv and iw.
 * The two int arguments fv and fw specify how many classes at the beginning
 * of each invariant are fixed on each side.
 *
 * This is a central function that is called often; so it should be as fast
 * as possible. Going to the next splitting involves copying one set of
 * invariants (iv,iw) to another, but we want to avoid making a 1:1 copy first
 * and then copy the next splitting back to the original variables.
 *
 * Therefore we pass an array ar of 4 pointers to inv's. All 4 inv's that the
 * pointers refer to must be different and have enough memory allocated. We
 * assume that the old splitting is in iv=*ar[0] and iw=*ar[1] when the
 * function is called. Afterwards the new splitting will be there too, but we
 * achieve this by writing the new splitting to *ar[2] and *ar[3] first, and
 * then exchanging the pointers.
 *
 * When the function is called, apart from the fixed classes all other
 * classes must be in iv=*ar[0], and the rational number must be 1. The moving
 * classes MUST be in sorted order. The non-changing entries (degree,
 * dimension, the fixed classes, ...) must be set correctly in all 4 inv's.
 *
 * After that, a statement like
 *   do {
 *     ...
 *   } while ( next_pair (...) );
 * will loop over all splittings of the classes. The rational number will
 * always be set to the corresponding binomial coefficient. The number dw
 * will always contain the virtual dimension of the iw invariant, provided
 * it was set correctly at the beginning (the function next_pair will only
 * make relative changes to this number when moving classes from one side
 * to the other). After all splittings have been returned, the function
 * returns FALSE and the two invariants are again in the same state as they
 * were at the beginning, i.e. all classes are in iv.
 *
 * Note: Only relative changes are made to coeff, so if there is a common
 * factor for all splittings, this can be achieved by setting coeff to this
 * value at the beginning.
 *
 */

bool next_pair (inv **ar, int fv, int fw, int& dw, rational& coeff)
{
	pdata *iv_src = &(ar[0]->pt[fv]);
	pdata *iw_src = &(ar[1]->pt[fw]);
	pdata *iv_dst = &(ar[2]->pt[fv]);
	pdata *iw_dst = &(ar[3]->pt[fw]);
	pdata *tmp;
	inv *ptr;

	int i,n;

	// If there are no classes left in iv, we're done.
	if (ar[0]->points==fv) return 0;

	// As long as the first class in iw is smaller than that in iv, copy
	// it (back) to iv. This doesn't introduce binomial coefficients as
	// blocks of same classes are always moved as a whole, but it changes
	// the virtual dimension.
	tmp = iw_src;
	for (i=0; i<ar[1]->points-fw && *tmp<*iv_src; i++, tmp++)
		dw += tmp->vdim(ar[0]->xdim)/2;
	// We want to move i points now...
	if (i) {
		memmove (iv_dst,iw_src,i*sizeof(pdata));
		iw_src += i;
		iv_dst += i;
	}

	// Now move _one_ class from iv to iw. Don't advance the pointers yet.
	dw -= iv_src->vdim(ar[0]->xdim)/2;
	*iw_dst = *iv_src;

	// Check if there are more of these classes. We'll leave them where
	// they are, but moving one of them will change the binomial
	// coefficient.
	tmp = iv_src; n = 1;
	while ( n<ar[0]->points-fv && *(++tmp)==*iw_dst ) n++;
	coeff *= n;
	tmp = iw_src; n = 1;
	while ( n<ar[1]->points-fw-i+1 && *(tmp++)==*iw_dst ) n++;
	coeff /= n;

	iw_dst++;
	iv_src++;

	// Copy the remaining classes unchanged. This will not change
	// the coefficient or the virtual dimension.
	if (ar[0]->points-fv-1)
		memmove (iv_dst,iv_src,(ar[0]->points-fv-1)*sizeof(pdata));
	if (ar[1]->points-fw-i)
		memmove (iw_dst,iw_src,(ar[1]->points-fw-i)*sizeof(pdata));

	ar[2]->points = ar[0]->points+i-1;
	ar[3]->points = ar[1]->points-i+1;

	// Finally, exchange ar[0], ar[1] with ar[2], ar[3].
	ptr = ar[0]; ar[0] = ar[2]; ar[2] = ptr;
	ptr = ar[1]; ar[1] = ar[3]; ar[3] = ptr;

	return 1;
}



// Compute a sum of invariants corresponding to all 2-component splittings.

rational split2_sum (inv& iv, inv& iw, int fv, int fw, int skip, int lvl)
{
	// Space for the 4 invariants needed by next_pair
	inv **ar;
	assert (ar = new inv*[4]);
	assert (ar[0] = new inv(iv));
	assert (ar[1] = new inv(iw));
	assert (ar[2] = new inv(iv));
	assert (ar[3] = new inv(iw));

	rational result = 0;
	rational coeff = rational(1)/iv.xdeg; // factor from diagonal class
	rational tmpv,tmpw;
	int dw;
	int d = iv.cdeg ; // the total degree
	int d1,d2; // the valid degree bounds in the current term for iw
	int i;
	int desc = 0; // 1 if there are descendants or we have negative genus

	if (iw.cgenus>=0) dw = (iv.xdim-3)*(1-iw.cgenus);
	else dw = iv.xdim-3+iw.cgenus;

	// Fill in the remaining initial values.
	ar[1]->xdim = ar[3]->xdim = iv.xdim;
	ar[1]->xdeg = ar[3]->xdeg = iv.xdeg;
	for (i=0; i<=3; i++)
		ar[i]->pt[0].psipow = ar[i]->pt[0].mult = 0;

	for (i=1; i<fw; i++) dw -= iw.pt[i].vdim (iv.xdim)/2;

	// Set 'desc'
	if (iv.pt[iv.points-1].psipow || (iv.cgenus<0) || (iw.cgenus<0)) {
		desc = 1;
	} else {
		for (i=0; i<fv; i++) if (iv.pt[i].psipow) desc=1;
		if (!desc)
		for (i=0; i<fw; i++) if (iw.pt[i].psipow) desc=1;
	}

	// Now do the loop over all class splittings...
	do {
		// ... and the loop over all degree splittings.
		// Compute the bounds for the degrees
		if (iv.xdim+2-iv.xdeg>0) {
			// the iv.xdim+1-iv.xdeg in the numerator is to
			// round up
			d1 = idiv (iv.xdim-iv.xdeg-dw,iv.xdim+2-iv.xdeg);
			if (d1<0) d1=0;
			d2 = idiv (iv.xdim-1-dw,iv.xdim+2-iv.xdeg);
			if (d2>d) d2=d;
		} else if (iv.xdim+2-iv.xdeg<0) {
			// the iv.xdeg-iv.xdim-3 in the numerator is to
			// round up
			d1 = idiv (dw-iv.xdim+iv.xdeg-iv.xdim-2,
				iv.xdeg-iv.xdim-2);
			if (d1<0) d1=0;
			d2 = idiv (1+dw,iv.xdeg-iv.xdim-2);
			if (d2>d) d2=d;
		} else {
			d1 = 0; d2 = d;
		}
		for (ar[1]->cdeg = d1; ar[1]->cdeg<=d2; ar[1]->cdeg++)
		if (--skip<0) {
			ar[0]->cdeg = d-ar[1]->cdeg;

			// fill in the diagonal class
			ar[0]->pt[0].Hpow = iv.xdim -
				(ar[1]->pt[0].Hpow =
					ar[1]->cdeg*(iv.xdim+2-iv.xdeg)+dw+1);
			// 1 in a non-descendant positive-degree class -> 0
			if (!ar[0]->pt[0].Hpow && ar[0]->cdeg && !desc)
				continue;
			if (!ar[1]->pt[0].Hpow && ar[1]->cdeg && !desc)
				continue;

			#ifdef SPLIT2_DEBUG
			if (ar[0]->pt[0].Hpow<0 || ar[1]->pt[0].Hpow<0)
				cout << "*split2* level " << lvl
				     << ": this shouldn't happen" << endl;
			#endif

			// Compute the two invariants. Take the one with
			// lowest genus first, and if both have the same
			// genus, the one with lowest degree. This one is
			// most likely to be known, and if it happens to be
			// 0 we don't have to compute the other one any more.
			#ifdef SPLIT2_DEBUG
			if (lvl<=SPLIT2_DEBUG)
				cout << "*split2* level " << lvl
				     << ": need term " << coeff << "*"
				     << *ar[0] << "*" << *ar[1] << endl;
			#endif
			if ((!iv.cgenus) || (iv.cgenus<iw.cgenus)
				|| (iv.cgenus==iw.cgenus &&
					ar[0]->cdeg<ar[1]->cdeg)) {
				// compute iv first
				tmpv = compute (*ar[0], lvl+1);
				if ( is_zero(tmpv) ) continue;
				tmpw = compute (*ar[1], lvl+1);
				if ( is_zero(tmpw) ) continue;
			} else {
				// compute iw first
				tmpw = compute (*ar[1], lvl+1);
				if ( is_zero(tmpw) ) continue;
				tmpv = compute (*ar[0], lvl+1);
				if ( is_zero(tmpv) ) continue;
			}
			result += coeff*tmpv*tmpw;
			#ifdef SPLIT2_DEBUG
			if (lvl<=SPLIT2_DEBUG)
				cout << "*split2* level " << lvl
				     << ": term is "
				     << coeff*tmpv*tmpw << endl;
			#endif
		}
		if ( is_undefined (result) ) break;
	} while ( next_pair (ar,fv,fw,dw,coeff) );

	delete ar[0];
	delete ar[1];
	delete ar[2];
	delete ar[3];
	delete ar;

	return result;
}



/*
 * The following four functions allow to reconstruct all genus-0 absolute
 * invariants from the corresponding 1-point invariants, for any variety:
 * 0) 1-point invariants are assumed to be known. (This is obvious for P^n.
 *    For other varieties, we'll use relative theory to compute these.)
 * 1) twopoint_onedesc() to compute from (0) the 2-point invariants with at
 *    most 1 descendant.
 * 2) twopoint_twodesc() to compute from (1) every 2-point invariant.
 * 3) first_reconstruction() to compute from (2) all invariants with no
 *    descendants.
 * 4) trr_genus0() to compute from (3) all invariants, including descendants.
 *
 * In all functions it is assumed that the invariant is sorted. The invariant
 * remains unchanged.
 *
 */



/*
 * Compute 2-point invariants with at most 1 descendant from 1-point
 * invariants. If we want to compute a 2-point invariant
 *   (H^k psi^l , H^a)
 * (with l possibly 0, but not a), we rewrite the invariant as
 *   (H^k psi^(l+1), H^a, 1)
 * using the fundamental class axiom. Then we split H^a as H*H^(a-1), i.e.
 * use WDVV with classes
 *   (1,H^k psi^(l+1) | H, H^(a-1)) <-> (H,H^k psi^(l+1) | 1, H^(a-1)).
 * This expresses the desired invariant in terms of others with
 * - smaller degree, or
 * - the same degree, and smaller a.
 * Note that the second splitting is almost trivial, because it contains a
 * non-descendant invariant with a fundamental class. This can only give a
 * non-zero contribution if the degree is 0. Hence, we don't call split2_sum
 * for this term, but rather compute the one single term directly.
 *
 * It is assumed that the invariant x is a rational, absolute 2-point
 * function with at most one descendant. As the classes are sorted, the
 * descendant is necessarily the second class.
 *
 */

rational twopoint_onedesc (inv& x, int lvl) {
	rational result;

	if (!x.pt[0].Hpow) return result; // a=0; this shouldn't happen as
		// otherwise we would have used the fundamental class axiom
		// before

	inv iv(3), iw(3);

	// Initializing
	iv.xdim = x.xdim;
	iv.xdeg = x.xdeg;
	iv.cgenus = iw.cgenus = 0;
	iv.cdeg = x.cdeg;
	iv.points = iw.points = 3;

	// The (trivial) second splitting
	iv.pt[0] = pdata (1);
	iv.pt[1] = x.pt[1];
	iv.pt[1].psipow++;
	iv.pt[2] = pdata (x.pt[0].Hpow-1);
	#ifdef SPLIT2_DEBUG
	if (lvl<=SPLIT2_DEBUG)
		cout << "*split2* level " << lvl
		     << ": compute initial term " << iv << endl;
	#endif
	result = compute (iv, lvl+1);
	#ifdef SPLIT2_DEBUG
	if (lvl<=SPLIT2_DEBUG)
		cout << "*split2* level " << lvl
		     << ": term is " << result << endl;
	#endif

	// The first splitting
	iv.pt[2] = pdata (0);
	iw.pt[1] = pdata (1);
	iw.pt[2] = pdata (x.pt[0].Hpow-1);
	result -= split2_sum (iv,iw,3,3,1,lvl);

	return result;
}



/*
 * Compute 2-point invariants with 2 descendants from 2-point invariants with
 * at most one descendant. If we want to compute a 2-point invariant
 *   (H^k psi^l , H^a psi^b)
 * ( with l and b non-zero), we use WDVV splitting with classes
 *   (1,H | H^k psi^l, H^a psi^b) <-> ... .
 * This expresses the desired invariant in terms of others with
 * - smaller degree, or
 * - the same degree, and lower psi power.
 * In fact, the first splitting is almost trivial, because it is a
 * non-descendant invariant with a fundamental class. This can only give a
 * non-zero contribution if the degree is 0, in which case it gives
 *   (H, H^k psi^l, H^a psi^b).
 * By the divisor axiom, this contains the desired invariant, plus correction
 * terms
 *   (H^(k+1) psi^(l-1), H^a psi^b) and (H^k psi^l, H^(a+1) psi^(b-1))
 * if they exist. Hence, we don't call split2_sum for this term, but rather
 * compute the two correction terms directly.
 *
 * It is assumed that the invariant x is a rational, absolute 2-point
 * function with two descendants.
 *
 */

rational twopoint_twodesc (inv& x, int lvl) {
	rational result = 0;
	rational tmp;

	inv iv(3); iv = x;
	inv iw(3);

	// The two divisor axiom corrections
	if (iv.pt[0].Hpow<NN) {
		iv.pt[0].Hpow++; iv.pt[0].psipow--;
		#ifdef SPLIT2_DEBUG
		if (lvl<=SPLIT2_DEBUG)
			cout << "*split2* level " << lvl
			     << ": compute 1st initial term " << iv << endl;
		#endif
		tmp = compute (iv, lvl+1);
		#ifdef SPLIT2_DEBUG
		if (lvl<=SPLIT2_DEBUG)
			cout << "*split2* level " << lvl
			     << ": term is " << tmp << endl;
		#endif
		iv.pt[0].Hpow--; iv.pt[0].psipow++;
		result = -tmp;
	}
	if (iv.pt[1].Hpow<NN) {
		iv.pt[1].Hpow++; iv.pt[1].psipow--;
		#ifdef SPLIT2_DEBUG
		if (lvl<=SPLIT2_DEBUG)
			cout << "*split2* level " << lvl
			     << ": compute 2nd initial term " << iv << endl;
		#endif
		tmp = compute (iv, lvl+1);
		#ifdef SPLIT2_DEBUG
		if (lvl<=SPLIT2_DEBUG)
			cout << "*split2* level " << lvl
			     << ": term is " << tmp << endl;
		#endif
		iv.pt[1].Hpow--; iv.pt[1].psipow++;
		result -= tmp;
	}

	// Initializing
	iw.cgenus = 0;
	iv.points = iw.points = 3;

	// The splitting
	iv.pt[2] = pdata (0);
	iw.pt[1] = x.pt[0];
	iw.pt[2] = pdata (1);
	result += split2_sum (iv,iw,3,3,0,lvl);

	return result/(DD*EE); // from the main term of the divisor axiom
}



/*
 * Compute invariants using the first reconstruction theorem of Kontsevich
 * and Manin. We assume that we have a g=0 absolute invariant with no
 * descendants, and that we have at least 3 marked points. We also assume
 * that none of the classes is a divisor or fundamental class.
 * Strategy: Suppose we want to compute an invariant with sorted
 * classes H^{i_1}, ..., H^{i_n}. Then use the WDVV equation with
 * special classes
 *   H, H^{i_1-1} | H^{i_{n-1}}, H^n  (LHS)
 *   H, H^n | H^{i_1-1}, H^{i_{n-1}}  (RHS)
 * This ensures that the invariant to compute appears, and that
 * all other invariants have
 * - smaller degree, or
 * - the same degree and smaller number of points (after applying
 *   the divisor axiom), or
 * - the same degree, the same number of points, but a bigger highest
 *   class (namely a H*H^n class).
 *
 */

rational first_reconstruction (inv& x, int lvl)
{
	rational result = 0;
	int i;

	if ( !x.pt[0].Hpow ) return rational ();

	// Create two temporary invariants with enough space
	inv iv(x.points);
	inv iw(x.points);

	// Initialize the invariants
	iv.xdim = x.xdim;
	iv.xdeg = x.xdeg;
	iv.cgenus = iw.cgenus = 0;
	iv.cdeg = x.cdeg;
	iv.points = x.points;
	iw.points = 3;

	for (i=1; i<x.points-2; i++) iv.pt[i+2] = x.pt[i];
	iv.pt[1] = x.pt[x.points-2];
	iv.pt[2] = x.pt[x.points-1];
	iw.pt[1] = pdata (1);
	iw.pt[2] = pdata (x.pt[0].Hpow-1);

	// left hand side of WDVV
	result = -split2_sum (iv,iw,3,3,1,lvl);

	iv.pt[1] = pdata (1);
	iw.pt[1] = x.pt[x.points-2];

	// right hand side of WDVV
	result += split2_sum (iv,iw,3,3,0,lvl);

	return result;
}



/*
 * Compute an invariant using the g=0 topological recursion relation
 * (see Getzler). We assume that there are at least 3 marked points,
 * and that there is at least one descendant class (at the last point).
 *
 */

rational trr_genus0 (inv& x, int lvl)
{
	int i;

	inv iv(x.points);
	inv iw(x.points);

	// Initialize the invariants
	iv.xdim = x.xdim;
	iv.xdeg = x.xdeg;
	iv.cgenus = iw.cgenus = 0;
	iv.cdeg = x.cdeg;
	iv.points = x.points;
	iw.points = 2;
	for (i=0; i<x.points-1; i++) iv.pt[i+1] = x.pt[i];
	iw.pt[1] = x.pt[x.points-1];
	iw.pt[1].psipow--;

	return split2_sum (iv,iw,3,2,0,lvl);
}



/*
 * Compute an elliptic invariant using the g=1 topological recursion relation
 * (see Getzler). pos has to be the point number of a descendant class. We
 * allow to specify this to be able to check the TRR relation with respect
 * to any marked point (currently we can't always compute elliptic invariants,
 * so it's worth trying every possibility). Later, when every elliptic
 * invariant can be computed, we may delete this argument and always take the
 * last class. We may also then delete the "if ( is_undefined ) return"
 * statements.
 *
 */

rational trr_genus1 (inv& x, int pos, int lvl)
{
	int i;

	rational result,tmp;

	inv iv(x.points+2);
	inv iw(x.points+1);

	// Initialize the invariants
	iv.xdim = x.xdim;
	iv.xdeg = x.xdeg;
	iv.cgenus = 0;
	iw.cgenus = 1;
	iv.cdeg = x.cdeg;
	iv.points = x.points+1;
	iw.points = 1;
	for (i=0; i<pos; i++) iv.pt[i+2] = x.pt[i];
	for (i=pos+1; i<x.points; i++) iv.pt[i+1] = x.pt[i];
	iv.pt[1] = x.pt[pos];
	iv.pt[1].psipow--;

	// The first term
	result = split2_sum (iv,iw,2,1,0,lvl);
	if ( is_undefined (result) ) return result;

	// The second (loop) term. Note that the genus is 0 here
	for (i=0; i<x.points; i++) iv.pt[i] = x.pt[i];
	iv.pt[pos].psipow--;
	iv.points = x.points+2;
	iv.pt[x.points] = pdata (0);
	iv.pt[x.points+1] = pdata (0);
	for (i=0; i<=iv.xdim; i++) {
		iv.pt[x.points].Hpow = i;
		iv.pt[x.points+1].Hpow = iv.xdim-i;
		#ifdef SPLIT2_DEBUG
		if (lvl<=SPLIT2_DEBUG)
			cout << "*split2* level " << lvl
			     << ": need term " << iv << endl;
		#endif
		tmp = compute (iv,lvl+1)/24;
		result += tmp;
		#ifdef SPLIT2_DEBUG
		if (lvl<=SPLIT2_DEBUG)
			cout << "*split2* level " << lvl
			     << ": term is "
			     << tmp << endl;
		#endif
		if ( is_undefined (result) ) return result;
	}

	return result;
}



// Check if we can compute an invariant by splitting it into two components.
// Return UNDEFINED if this doesn't work. In this case, return the invariant
// unchanged. If the computation succeeds, we store the result in the database.

rational split2 (inv& iv, int lvl)
{
	rational result;
	int i;

	// Relative invariants are not supported here.
	if ( iv.points && iv.pt[iv.points-1].mult ) return rational ();

	// Neither are negative genus invariants.
	if ( GG<0 ) return rational ();

	// Currently, invariants with primitive classes are not supported
	// either.
	if ( (EE>1) && iv.reduce() ) return rational ();

	// Compute 2-point rational absolute invariants with at most one
	// descendant.
	if ( !GG && iv.points==2 && !iv.pt[0].psipow ) {
		#ifdef SPLIT2_DEBUG
		DEBUG_BEFORE ("WDVV for 2-point invariants with at most " \
				"1 descendant");
		#endif
		result = twopoint_onedesc (iv,lvl);
		#ifdef SPLIT2_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// Compute 2-point rational absolute invariants with two
	// descendants.
	if ( !GG && iv.points==2 && iv.pt[0].psipow ) {
		#ifdef SPLIT2_DEBUG
		DEBUG_BEFORE ("WDVV for 2-point invariants with 2 " \
				"descendants");
		#endif
		result = twopoint_twodesc (iv,lvl);
		#ifdef SPLIT2_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// Compute any n-point rational absolute invariant without descendants.
	if ( !GG && iv.points>2 && !iv.pt[iv.points-1].psipow ) {
		#ifdef SPLIT2_DEBUG
		DEBUG_BEFORE ("first reconstruction theorem");
		#endif
		result = first_reconstruction (iv,lvl);
		#ifdef SPLIT2_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// Compute any n-point rational absolute invariant with descendants
	// using the g=0 topological recursion relation.
	if ( !GG && iv.points>2 && iv.pt[iv.points-1].psipow ) {
		#ifdef SPLIT2_DEBUG
		DEBUG_BEFORE ("g=0 topological recursion relation");
		#endif
		result = trr_genus0 (iv,lvl);
		#ifdef SPLIT2_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// Try to compute an elliptic invariant using the g=1 topological
	// recursion relations. This is currently not supported for
	// hypersurfaces of degree>1 because of primitive class complications
	// in the loop term.
	if ( GG==1 && EE==1 ) for (i=iv.points-1; i>=0; i--) {
		if ( !iv.pt[i].psipow ) break;
		#ifdef SPLIT2_DEBUG
		DEBUG_BEFORE ("g=1 topological recursion relation");
		#endif
		result = trr_genus1 (iv,i,lvl);
		#ifdef SPLIT2_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// Sorry, but we can't compute the invariant here. Exit.
	return rational ();

STORE:	// We've successfully computed the invariant. Store it in the database.
	GWdbm.reset_key ();
	GWdbm.reset_value ();
	set_key (GWdbm, iv);
	set_value (GWdbm, result);
	GWdbm.write ();
	return result;
}

