/*
 * rational.cc - a class for rational number arithmetic
 *
 * The implementation is very straightforward. No surprises here...
 *
 */

#include <assert.h>

#include "rational.h"

/*
 * If we include a file defining RATIONAL_DEBUG, debugging output is printed
 * whenever the reference counter of a number is raised or dropped. The last
 * message printed at program exit should be that the local and global
 * reference counters are both 0.
 *
 */

#include "config.h"




void rational_data::rational_debug (mpq_t num, int n, int d)
{
	#ifdef RATIONAL_DEBUG

	static int totalrefcnt = 0;

	totalrefcnt += d;
	char *str;
	gmp_asprintf (&str,"%Qd",num);

	cout << "*rational* value " << str;
	cout << ", local ref count " << n;
	cout << ", global ref count " << totalrefcnt << endl;

	assert (n>=0 && totalrefcnt>=0);

	free (str);

	#endif
}



// We need an mpq constant 1 and don't want to define it over and over again.

static mpq_t mpq_one;
int mpq_one_defined = 0;

inline void define_one ()
{
	if (!mpq_one_defined) {
		mpq_init (mpq_one);
		mpq_set_si (mpq_one,1,1);
		mpq_one_defined = 1;
	}
}



// Why is there no power function for mpq in GMP?

void my_mpq_pow (mpq_t res, mpq_t x, int n)
{
	int e = (n>=0 ? n : -n);

	mpz_pow_ui (mpq_numref (res), mpq_numref (x),e);
	mpz_pow_ui (mpq_denref (res), mpq_denref (x),e);

	if (n<0) mpq_inv (res,res);
}



#define RATIONAL_OP2(FUNC,COND,THIRD)					\
{									\
	rational result;						\
									\
	if (!x.ptr COND) return result; /* undefined */			\
									\
	assert (result.ptr = new rational_data);			\
	FUNC (result.ptr->number, x.ptr->number, THIRD);		\
									\
	return result;							\
}

rational operator+ (const rational& x, const rational& y)
	RATIONAL_OP2 (mpq_add,|| !y.ptr,y.ptr->number);
rational operator- (const rational& x, const rational& y)
	RATIONAL_OP2 (mpq_sub,|| !y.ptr,y.ptr->number);
rational operator* (const rational& x, const rational& y)
	RATIONAL_OP2 (mpq_mul,|| !y.ptr,y.ptr->number);
rational operator/ (const rational& x, const rational& y)
	RATIONAL_OP2 (mpq_div,|| !y.ptr || !mpq_sgn (y.ptr->number),\
		y.ptr->number);
rational operator<< (const rational& x, const unsigned& n)
	RATIONAL_OP2 (mpq_mul_2exp,,n);
rational operator>> (const rational& x, const unsigned& n)
	RATIONAL_OP2 (mpq_div_2exp,,n);
rational pow (const rational& x, const int& n)
	RATIONAL_OP2 (my_mpq_pow,|| (n<0 && !mpq_sgn (x.ptr->number)),n);



#define RATIONAL_OP1(FUNC,COND,THIRD)					\
{									\
	rational_data *ptr2 = ptr;					\
									\
	if (!ptr) return *this; /* first argument undefined */		\
									\
	if (COND) { /* second argument undefined, */			\
			   /* so erase first one too */			\
		if ( ! --(*ptr) ) delete ptr;				\
		ptr = NULL;						\
		return *this;						\
	}								\
									\
	if ( ptr->cnt()>1 ) {						\
		/* We're about to change the value of the current */	\
		/* number. If more than one pointer refers to it, */	\
		/* we have to make a copy... */				\
		--(*ptr);						\
		assert (ptr = new rational_data);			\
	}								\
									\
	/* ... otherwise we can overwrite the value. */			\
	FUNC (ptr->number,ptr2->number,THIRD);				\
	return *this;							\
}

rational& rational::operator+= (const rational& x)
	RATIONAL_OP1 (mpq_add,!x.ptr,x.ptr->number);
rational& rational::operator-= (const rational& x)
	RATIONAL_OP1 (mpq_sub,!x.ptr,x.ptr->number);
rational& rational::operator*= (const rational& x)
	RATIONAL_OP1 (mpq_mul,!x.ptr,x.ptr->number);
rational& rational::operator/= (const rational& x)
	RATIONAL_OP1 (mpq_div,!x.ptr || !mpq_sgn (x.ptr->number),\
		x.ptr->number);
rational& rational::operator<<= (const unsigned& n)
	RATIONAL_OP1 (mpq_mul_2exp,0,n);
rational& rational::operator>>= (const unsigned& n)
	RATIONAL_OP1 (mpq_div_2exp,0,n);



rational operator- (const rational& x)
{
	rational result;

	if (!x.ptr) return result;

	assert (result.ptr = new rational_data);
	mpq_neg (result.ptr->number, x.ptr->number);

	return result;
}

rational operator+ (const rational& x, const int& y) {return x+rational(y);};
rational operator+ (const int& x, const rational& y) {return rational(x)+y;};
rational operator- (const rational& x, const int& y) {return x-rational(y);};
rational operator- (const int& x, const rational& y) {return rational(x)-y;};
rational operator* (const rational& x, const int& y) {return x*rational(y);};
rational operator* (const int& x, const rational& y) {return rational(x)*y;};
rational operator/ (const rational& x, const int& y) {return x/rational(y);};
rational operator/ (const int& x, const rational& y) {return rational(x)/y;};
rational pow (const int& x, const int& n) {return pow (rational(x),n);};
rational operator+ (const rational& x) {return x;};



// Comparisons

#define RATIONAL_CMP(FUNC,COND) (const rational& x, const rational& y)	\
{									\
	if (!x.ptr || !y.ptr) return 0;					\
									\
	return FUNC(x.ptr->number,y.ptr->number) COND;			\
}

bool operator== RATIONAL_CMP (mpq_equal,);
bool operator!= RATIONAL_CMP (mpq_equal,==0);
bool operator< RATIONAL_CMP (mpq_cmp,<0);
bool operator<= RATIONAL_CMP (mpq_cmp,<=0);
bool operator> RATIONAL_CMP (mpq_cmp,>0);
bool operator>= RATIONAL_CMP (mpq_cmp,>=0);

bool operator== (const rational& x, const int& y) {return x==rational(y);};
bool operator== (const int& x, const rational& y) {return rational(x)==y;};
bool operator!= (const rational& x, const int& y) {return x!=rational(y);};
bool operator!= (const int& x, const rational& y) {return rational(x)!=y;};
bool operator< (const rational& x, const int& y) {return x<rational(y);};
bool operator< (const int& x, const rational& y) {return rational(x)<y;};
bool operator<= (const rational& x, const int& y) {return x<=rational(y);};
bool operator<= (const int& x, const rational& y) {return rational(x)<=y;};
bool operator> (const rational& x, const int& y) {return x>rational(y);};
bool operator> (const int& x, const rational& y) {return rational(x)>y;};
bool operator>= (const rational& x, const int& y) {return x>=rational(y);};
bool operator>= (const int& x, const rational& y) {return rational(x)==y;};



// Casting

bool is_undefined (const rational& x) {return !x.ptr;};

int to_int (const rational& x) {
	if (!x.ptr) return 0;
	return mpz_get_si (mpq_numref (x.ptr->number));
};
bool is_zero (const rational& x) {
	return x.ptr && ! mpq_sgn (x.ptr->number);
};



bool is_integer (const rational& x)
{
	if (!x.ptr) return 0;

	define_one ();
	return !mpz_cmp (mpq_denref (x.ptr->number),mpq_denref (mpq_one));
}

bool is_int (const rational& x)
{
	if (!x.ptr) return 0;

	define_one ();

	return !mpz_cmp (mpq_denref(x.ptr->number),mpq_denref(mpq_one))
		&& mpz_fits_sint_p (mpq_numref (x.ptr->number));
}



rational divconv (int n) {return rational(n);};
rational divconv (rational& x) {return x;};



ostream& operator<< (ostream& os, const rational& x)
{
	if (x.ptr) {
		char *str;
		gmp_asprintf (&str,"%Qd",x.ptr->number);
		os << str;
		free (str);
	} else os << "UNDEFINED";

	return os;
}



/* Input a rational number. The number doesn't have to be canonicalized.
 *
 * We store the input in a string first, and then call the corresponding
 * sscanf function. A '/' will be read only once, and only if it's followed
 * by a digit. If there is no valid input, the input stream will be set to
 * bad state, and the resulting number is UNDEFINED. Whitespace is not allowed
 * within the number (i.e. after the '-' or around the '/').
 *
 */

istream& operator>> (istream& is, rational& x)
{
	char *str = NULL;
	char *str2;
	char c,d;
	int i;
	int slash = 0;
	int alloc = 0; // memory allocated
	int read = 0; // bytes read

	// Skip whitespace
	do {
		c = is.get ();
		if (!is) goto BAD;
	} while (c==' ' || c=='\n' || c=='\t');
	is.unget ();

	do {
		if (is.eof ()) break;
		c = is.get ();
		if (!is) goto BAD;

		// Store character
		if (read>=alloc-3) {
			// allocate more memory
			str2 = str;
			alloc += 32;
			str = (char *)malloc (alloc);
			assert (str);
			for (i=0; i<read; i++) str[i] = str2[i];
			free (str2);
		}
		str[read++] = c;

		if (c>='0' && c<='9') continue; // numbers are always OK
		if (c=='-' && read==1) continue; // '-' OK if it comes first
		if (c=='/' && !slash && !is.eof ()) {
			slash = 1; // '/' OK once if followed by a number
			c = is.get ();
			if (!is) goto BAD;
			is.unget ();
			if (c>='0' && c<='9') continue;
		}

		is.unget (); // end of number, last character back to is
		break;
	} while (1);

	str[--read] = ','; // end of string
	str[++read] = '0'; // add a second number 0
	str[++read] = 0;
	if (!read) goto BAD; // no input

	if (x.ptr) if ( ! --(*x.ptr) ) delete x.ptr;
	assert (x.ptr = new rational_data);
	if ( gmp_sscanf (str,"%Qd,%d",x.ptr->number,&i) != 2 ) goto BAD;

	if (!mpz_sgn (mpq_denref (x.ptr->number))) goto BAD; // division by 0
	mpq_canonicalize (x.ptr->number);

	return is;

BAD:	is.clear ( ios::badbit | is.rdstate() );
	if (x.ptr) if ( ! --(*x.ptr) ) delete x.ptr;
	x.ptr = NULL;
	free (str);
	return is;
}



// Fast factorial function. We store the values in an array for fast access.

class fact_class {
	public:
		rational *value; // We store 16 values here.
		fact_class *next;

		fact_class (rational start, int nstart) {
			assert (value = new rational[16]);
			// Compute the next 16 values of the factorial
			// function.
			for (int i=0; i<=15; i++) {
				++nstart;
				start *= (nstart>0 ? nstart : 1);
				value[i] = start;
			}
			next = NULL;
		};
		// We never delete a fact_class instance, so we don't bother
		// about destructors.
};

fact_class fact_data(1,-1);

rational factorial (int n)
{
	int k=-1;

	if (n<0) return rational();

	fact_class *ptr = &fact_data;
	while (n>=16) {
		k += 16;
		if (!ptr->next)
			assert (ptr->next = new fact_class(ptr->value[15],k) );
		ptr = ptr->next;
		n -=16;
	}

	return ptr->value[n];
}



// **********   Database functions   **********
//
// We provide functions to store rational numbers in a database. The algorithm
// to convert rational numbers to a sequence of bytes is similar to that used
// for integers in database.cc . It is optimized to use as little space as
// possible for numbers with small numerators and denominators.
//
// More precisely, the numerator is stored as a sequence of bits as follows:
//  a) 1 1 ... 1 1 0   with n 1's, if n+1 is the total number of bytes for
//       the numerator
//  b) 1 bit for the sign
//  c) 1 bit flag: is there a denominator (!=1) ?
//  d) data bits of length determined by n, with the most significant bytes
//       first.
// If the flag (c) indicates so, the denominator is stored after that in the
// same way, just with (b) and (c) left out.

// How many dual digits does the mpz x have? For the number 0, return 1.

static int numbits (mpz_t x)
{
	int n = mpz_size (x);
	if (!n) return 1; // the number is 0

	// Get the most significant limb
	mp_limb_t li = mpz_getlimbn (x,n-1);
	n = 8*(n-1)*sizeof(mp_limb_t)+1;
	while (li & mp_limb_t(-256)) {li>>=8; n+=8;};
	if (li & mp_limb_t(-16)) {li>>=4; n+=4;};
	if (li & mp_limb_t(-4)) {li>>=2; n+=2;};
	if (li & mp_limb_t(-2)) n+=1;

	return n;
}



// Make set_key/value_bits a nonmember function and define pointers to it

static inline void set_key_bits (database& db, int n, int d)
{
	db.set_key_bits (n,d);
}

static inline void set_value_bits (database& db, int n, int d)
{
	db.set_value_bits (n,d);
}

typedef void (*set_bits_func)(database&, int, int);

static set_bits_func SKEYBITS = &set_key_bits;
static set_bits_func SVALUEBITS = &set_value_bits;



// Using the function func (set_key_bits or set_value_bits), write the mpz
// number z and the nb least significant bits of b.

void set_mpz_bit (database& db, set_bits_func func, mpz_t x, int b, int nb)
{
	mp_limb_t li;
	int bytes = (numbits (x)+nb+6)/7; // the number of bytes we need
	int i,j,n,nli;

	// Write the leading 1 1 ... 1 1 0 sequence
	n = bytes;
	while (n>8) {
		(*func) (db,255,8);
		n-=8;
	}
	(*func) (db,(1<<n)-2,n);

	// Write the special bits
	(*func) (db, b, nb);

	// Write the mpz number
	n = 7*bytes-nb; // the number of bits we have to write
	nli = n & (8*sizeof(mp_limb_t)-1); // bits in the highst limb
	if (nli) {
		// there is a partial limb to write. Do so.
		li = mpz_getlimbn (x,(n-1)/(8*sizeof(mp_limb_t)));
		// Write the partial part of the highest byte
		if (i = (nli & 7)) {
			nli -= i;
			(*func) (db, li>>nli, i);
		}
		// More bytes in the highest limb
		while (nli) {
			nli -= 8;
			(*func) (db, li>>nli, 8);
		}
	}

	// Write the remaining complete limbs
	for (i= n / (8*sizeof(mp_limb_t)) - 1 ; i>=0; i--) {
		li = mpz_getlimbn (x,i);
		for (j=8*(sizeof(mp_limb_t)-1); j>=0; j-=8) {
			(*func) (db, (li>>j) & 255, 8);
		}
	}
}



void set_key (database& db, const rational& x)
{
	if (!x.ptr) {
		db.set_key_bits (64,8); // undefined, corresponds to "-0"
		return;
	}

	define_one ();

	int b = (mpz_sgn(mpq_numref(x.ptr->number))<0 ? 2 : 0);
	if (mpz_cmp (mpq_denref(x.ptr->number),mpq_denref(mpq_one))) b+=1;
	set_mpz_bit (db,SKEYBITS,mpq_numref(x.ptr->number),b,2);
	if (b&1) set_mpz_bit (db,SKEYBITS,mpq_denref(x.ptr->number),0,0);
}



void set_value (database& db, const rational& x)
{
	if (!x.ptr) {
		db.set_value_bits (64,8); // undefined, corresponds to "-0"
		return;
	}

	define_one ();

	int b = (mpz_sgn(mpq_numref(x.ptr->number))<0 ? 2 : 0);
	if (mpz_cmp (mpq_denref(x.ptr->number),mpq_denref(mpq_one))) b+=1;
	set_mpz_bit (db,SVALUEBITS,mpq_numref(x.ptr->number),b,2);
	if (b&1) set_mpz_bit (db,SVALUEBITS,mpq_denref(x.ptr->number),0,0);
}



// Get the mpz number z and nb bits in b.

bool get_mpz_bit (database& db, mpz_t x, int&b, int nb)
{
	char *s;
	int i,n,nli,t;

	// Get the number of bytes
	int bytes = 0;
	if ( ! db.get_value_bytes (s,0) ) return 0;
	while ( (*s)==char(255) ) {
		bytes += 8;
		if ( ! db.get_value_bytes (s,1) ) return 0;
		++s;
	}
	do {
		if ( ! db.get_value_bits (i,1) ) return 0;
		bytes++;
	} while (i);

	// Read the special bits
	if (nb) if ( ! db.get_value_bits (b,nb) ) return 0;

	// Read the remaining bits.
	n = 7*bytes-nb; // the number of bits to be read. This cannot be 0.
	nli = (n+7)>>3;
	if (n&7) {
		if ( ! db.get_value_bits (t,n&7) ) return 0;
		if ( ! db.get_value_bytes (s,n>>3) ) return 0;
		s--;
		s[0]=(char)t;
	} else {
		if ( ! db.get_value_bytes (s,n>>3) ) return 0;
	}
	mpz_import (x,nli,1,1,0,0,(void*)s);

	return 1;
}



bool get_value (database& db, rational& x)
{
	int b,c;

	x = 0; // make sure x is initialized and the denomiator is 1

	if ( ! get_mpz_bit (db, mpq_numref(x.ptr->number), b, 2) ) return 0;
	if ( b&1 )
	if ( ! get_mpz_bit (db, mpq_denref(x.ptr->number), c, 0) ) return 0;
	if (b&2) {
		// minus sign
		if (x.ptr->number->_mp_num._mp_size)
//			x.ptr->number->_mp_num._mp_size *= -1;
			x = -x;
		else x = rational (); // -0 -> undefined
	}

	return 1;
}

